package de.hftstuttgart.dtabackend.utils;

import com.github.dockerjava.api.DockerClient;
import com.github.dockerjava.api.async.ResultCallback;
import com.github.dockerjava.api.command.CreateContainerResponse;
import com.github.dockerjava.api.exception.DockerException;
import com.github.dockerjava.api.model.Bind;
import com.github.dockerjava.api.model.Frame;
import com.github.dockerjava.api.model.HostConfig;
import com.github.dockerjava.api.model.Mount;
import com.github.dockerjava.api.model.MountType;
import com.github.dockerjava.api.model.Volume;
import com.github.dockerjava.core.DefaultDockerClientConfig;
import com.github.dockerjava.core.DockerClientConfig;
import com.github.dockerjava.core.DockerClientImpl;
import com.github.dockerjava.zerodep.ZerodepDockerHttpClient;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.UUID;
import java.util.stream.Collectors;

import java.nio.file.Path;
import java.nio.file.Paths;


@Component
public class DockerUtil {
    private static final Logger LOG = LogManager.getLogger(DockerUtil.class);

    private DockerClient dockerClient;

    public DockerUtil(Environment env) {
        LOG.info("initializing Docker tools");
        LOG.debug("create docker client config");
        DockerClientConfig dockerClientConfig = DefaultDockerClientConfig.createDefaultConfigBuilder().build();

        LOG.debug("create docker http client");
        ZerodepDockerHttpClient httpClient = new ZerodepDockerHttpClient.Builder()
                .dockerHost(dockerClientConfig.getDockerHost())
                .sslConfig(dockerClientConfig.getSSLConfig())
        		.build();

        LOG.debug("create docker client");
        dockerClient = DockerClientImpl.getInstance(dockerClientConfig, httpClient);
    }

    public int runContainer(String image, Path testPath, Path srcPath, Path resultPath) 
        throws InterruptedException, IOException {
    LOG.debug(String.format("Pulling image: %s", image));
    try {
        dockerClient.pullImageCmd(image)
            .start()
            .awaitCompletion()
            .close();
    } catch (DockerException e) {
        LOG.error(String.format(
            "Pulling Docker image %s failed with %s, trying with local image",
            image, e.getMessage()));
    }

    LOG.debug("Creating container");
    CreateContainerResponse containerResponse;
    try {
        // Validate paths are absolute
        String absTestPath = ensureAbsolutePath(testPath);
        String absSrcPath = ensureAbsolutePath(srcPath);
        String absResultPath = ensureAbsolutePath(resultPath);

        // Debugging the paths
        LOG.debug(String.format(
            "Volume bindings: %s -> /data/test, %s -> /data/src, %s -> /data/result",
            absTestPath, absSrcPath, absResultPath));

        // Create the container
        containerResponse = dockerClient.createContainerCmd("testcontainer")
            .withImage(image)
            .withHostConfig(
                HostConfig.newHostConfig()
                    .withBinds(
                        new Bind(absTestPath, new Volume("/data/test")),
                        new Bind(absSrcPath, new Volume("/data/src")),
                        new Bind(absResultPath, new Volume("/data/result"))
                    )
            )
            .withCmd(
                "java",
                "-Djava.security.egd=file:/dev/./urandom",
                "-jar",
                "/data/app.jar",
                "/data/src:/data/test",
                "/data/libs/*:/data/test/libs/*",
                "/data/result"
            )
            .exec();
    } catch (DockerException e) {
        LOG.error(String.format("Creating Docker Testrunner container failed with %s", e.getMessage()));
        throw e;
    }

    LOG.debug(String.format("Container created: %s", containerResponse.getId()));

    LOG.debug(String.format("Starting container %s", containerResponse.getId()));
    dockerClient.startContainerCmd(containerResponse.getId()).exec();

    LOG.debug(String.format("Retrieving logs for container %s", containerResponse.getId()));
    dockerClient.logContainerCmd(containerResponse.getId())
        .withStdOut(true)
        .withStdErr(true)
        .withFollowStream(true)
        .exec(new ResultCallback.Adapter<Frame>() {
            @Override
            public void onNext(Frame frame) {
                LOG.debug(String.format("LOG: %s", new String(frame.getPayload(), StandardCharsets.UTF_8)));
            }
        });

    LOG.debug(String.format("Waiting for completion of container %s", containerResponse.getId()));
    int ret = dockerClient
        .waitContainerCmd(containerResponse.getId())
        .start()
        .awaitCompletion()
        .awaitStatusCode();
    LOG.debug(String.format("Container completed with status %d", ret));

    LOG.debug(String.format("Deleting container %s", containerResponse.getId()));
    dockerClient.removeContainerCmd(containerResponse.getId())
        .withRemoveVolumes(true)
        .exec();

    return ret;
}

private String ensureAbsolutePath(Path path) {
    if (!path.isAbsolute()) {
        throw new IllegalArgumentException("Path must be absolute: " + path.toString());
    }
    return path.toString();
}

}
